import os
import os.path
import sys
import time
import subprocess
import cairo
import gtk
import gtk.glade
import gobject
import dbus
import gettext
import math
import pwd
from gettext import gettext as _
from pkg_resources import resource_filename
from backup_manager.core import *

class BackupManagerLogWindow ():
    def __init__ (self):
        self._widgets = gtk.glade.XML (resource_filename (__name__, "backup.glade"))
        self._window = self._widgets.get_widget ("log_window")
        self._close_button = self._widgets.get_widget ("log_close_button")
        self._view = self._widgets.get_widget ("log_view")
        self._buffer = gtk.TextBuffer ()
        self._view.set_buffer (self._buffer)
        self._running = True
        self._worker = None
        
        self._window.connect ("delete-event", lambda w, e: self._window.hide () or True)
        self._close_button.connect ("clicked", lambda b: self._window.hide () or True)

    def start (self):
        if not self._worker:
            self._worker = BackupManagerThread (target=self._watch_log)
            self._worker.start ()

    def _watch_log (self):
        lf = get_log_file_location ()
        fd = None

        while self._running:
            if not fd:
                if not os.path.exists (lf):
                    time.sleep (3)
                    continue
                else:
                    fd = os.open (lf, os.O_RDONLY)

            data = os.read (fd, 1024)
            if data != "":
                gtk.gdk.threads_enter ()
                self._buffer.insert (self._buffer.get_end_iter (), data)
                gtk.gdk.threads_leave ()
            else:
                time.sleep (3)

        if fd:
            os.close (fd)
    
    def stop (self):
        self._running = False

    def show_all (self):
        self._window.present ()
        self._window.show_all ()

class BackupManagerApplet (gtk.StatusIcon):
    def __init__ (self):
        gtk.StatusIcon.__init__ (self)

        self.set_tooltip (_("Your data is being backed up.  Please don't reboot or shutdown."))
        self._menu = gtk.Menu ()
        self._quit_item = gtk.ImageMenuItem (stock_id=gtk.STOCK_QUIT)
        self._quit_item.show_all ()
        self._menu.append (self._quit_item)
        self._log_window = BackupManagerLogWindow ()
        self._showing = False
        self._manager_found = False
        self._timestamp = time.time ()
        
        self._sections = 10 
        aps = (2 * math.pi) / self._sections
        self._spacing_size =  aps * .1
        self._section_size = aps - self._spacing_size
        self._opacity_scale = 1.0 / float (self._sections)
        self._pixbufs = []
        self._idx = 0

        self._color = gtk.Entry ().style.bg[gtk.STATE_SELECTED]
        self._create_all ()
        self._quit_item.connect ("activate", self._quit)
        self.connect ("popup-menu", self._popup_menu)
        self.connect ("activate", self._activate)
        
        gobject.timeout_add (3000, self.backup_manager_running)
        gobject.timeout_add (200, self._animate_cb)

    def _quit (self, b):
        self._log_window.stop ()
        gtk.main_quit ()

    def _popup_menu (self, icon, button, time):
        self._menu.popup (None, None, None, button, time)

    def _activate (self, icon):
        self._log_window.show_all ()
        if not self._showing:
            self._showing = True
            self._log_window.start ()

    def _animate_cb (self):
        self.set_from_pixbuf (self._pixbufs[self._idx])
        self._idx = (self._idx + 1) % self._sections
        return True
        
    def _create_all (self):
        w = h = 24
        png = os.path.join (os.path.dirname (get_config_location ()), "tmp.png")
        png_dir = os.path.dirname (png)

        if not os.path.exists (png_dir):
            os.makedirs (png_dir, mode=0755)
            os.chmod (png_dir, 0755)
        
        for p in xrange (0, self._sections):
            imgs = cairo.ImageSurface (cairo.FORMAT_ARGB32, w, h)
            cr = cairo.Context (imgs)
            cr.set_line_width (5)
            spot = 0.0
            
            for i in xrange (0, self._sections):
                cr.new_path ()
                
                ri = i
                rd = 0
                for j in xrange (0, self._sections):
                    if ri == p:
                        break
                
                    ri = (ri + 1) % self._sections
                    rd += 1

                spot = spot + self._section_size + self._spacing_size
                cr.set_source_rgba (self._color.red / 65535.0,
                                    self._color.green / 65535.0,
                                    self._color.blue / 65535.0,
                                    (self._sections - rd) * self._opacity_scale)
                cr.arc (w / 2, h / 2, h / 2 - 5, spot, spot + self._section_size)
                cr.stroke ()
            
            imgs.write_to_png (png)
            self._pixbufs.append (gtk.gdk.pixbuf_new_from_file (png))
            os.unlink (png)
    
    def backup_manager_running (self):
        p = subprocess.Popen (["ps x --user %s | grep backup-manager |  grep -v grep | awk '{print $1}'" % \
                               pwd.getpwuid (os.getuid ())[0]], stdout=subprocess.PIPE, close_fds=True, shell=True)
        pids = p.stdout.read ().strip ()
        p.wait ()

        if pids != "":
            self._manager_found = True 
        else:
            if self._manager_found:
                print _("Manager isn't running anymore.  Exiting...")
                self._quit (None)
                return False
            else: 
                if time.time () - self._timestamp > 3000:
                    print _("Haven't found the manager in 5 minutes.  Exiting...")
                    self._quit (None)
                    return False

        return True

def main ():
    gtk.gdk.threads_init ()

    gettext.bindtextdomain ("backup-manager", "/usr/share/locale")
    gettext.textdomain ("backup-manager")

    bus = dbus.SessionBus ()
    ret = bus.request_name ("org.opensuse.BackupManagerApplet", dbus.bus.NAME_FLAG_DO_NOT_QUEUE)
    if ret != dbus.bus.REQUEST_NAME_REPLY_PRIMARY_OWNER:
        print _("An instance is already running...")
        sys.exit (0)
    
    bma = BackupManagerApplet ()
    gtk.main ()

if __name__ == "__main__":
    log = get_logger (redirect=True)
    
    try:
        main ()
    except KeyboardInterrupt:
        sys.exit (0)
    except Exception, e:
        log.error (str (e))
        sys.exit (1)
