import sys
import os
import pwd
import time
import string
import subprocess
import json
import errno
import dbus
import gobject
import threading
import logging
from optparse import OptionParser
from statvfs import *
from gettext import gettext as _
from optparse import OptionParser
from pkg_resources import resource_filename

SCHEDULE_NEVER = 0
SCHEDULE_HOURLY = 1
SCHEDULE_DAILY = 2
SCHEDULE_WEEKLY = 3
SCHEDULE_MONTHLY = 4

PURGE_DAYS = 0
PURGE_WEEKS = 1
PURGE_MONTHS = 2

schedule_map = { SCHEDULE_NEVER: _("Never"), 
                 SCHEDULE_HOURLY: _("Hourly"),
                 SCHEDULE_DAILY: _("Daily"),
                 SCHEDULE_WEEKLY: _("Weekly"), 
                 SCHEDULE_MONTHLY:  _("Monthly") }

purge_map = { PURGE_DAYS: _("Day(s)"),
              PURGE_WEEKS: _("Week(s)"),
              PURGE_MONTHS: _("Month(s)") }

logger_obj = None

class BackupManagerException (Exception): pass

class IBackupManager:
    def has_capability (self, cap):
        caps = self.get_capabilities ()
        return cap in caps if caps != None else None
    def get_capabilities (self): pass
    def verify_config (self, config): pass
    def verify_restore_data (self, src, dest, files, time): pass
    def backup (self, config): pass
    def restore (self, src, dest, time): pass
    def restore_specific (self, src, dest, files_tpl_lst): pass
    def restore_search (self, config, key): pass
    def purge_old_data (self, config): pass

class Config:
    def __init__ (self):
        self.config_file = None

    def __getattr__ (self, name):
        return None

    def __str__ (self):
        ret = ""
        for k, v in self.__dict__.iteritems ():
            ret += "%s: %s\n" % (k, v) 
        return ret

    def remove (self, name):
        if self.__dict__.has_key (name):
            del self.__dict__[name]

    def load (self, conf):
        fo = file (conf, 'r')
        self.__dict__ = json.loads (fo.read ())
        fo.close ()
        self.config_file = conf

    def save (self, conf=None):
        if not conf:
            conf = self.config_file

        print self
        conf_dir = os.path.dirname (conf) 
        if not os.path.exists (conf_dir):
            os.makedirs (conf_dir, mode=0755)
            os.chmod (conf_dir, 0755)

        fo = file (conf, 'w')
        fo.write (json.dumps (self.__dict__))
        fo.close ()

class BackupManagerThread (threading.Thread):
    def __init__ (self, *args, **kwargs):
        threading.Thread.__init__ (self, *args, **kwargs)
        self._finish_cb = None
        self._error_cb = None

    def set_callbacks (self, finish_cb=None, error_cb=None):
        self._finish_cb = finish_cb
        self._error_cb = error_cb

    def get_name (self):
        return self.name

    @staticmethod
    def join_thread (t):
        t.join ()
        return False
    
    def run (self):
        try:
            threading.Thread.run (self)
        except BackupManagerException, e:
            print >> sys.stderr, str (e)
            if self._error_cb:
                self._error_cb (str (e))
        else:
            if self._finish_cb:
                self._finish_cb ()
        
        gobject.idle_add (BackupManagerThread.join_thread, self)

def get_env_from_pid (pid):
    ret = {}
    fo = file ("/proc/%s/environ" % (pid), "r")
    vars = fo.read ()
    for var in vars.split ("\0"):
        try:
            k, v = var.split ("=")
        except ValueError:
            continue

        ret[k] = v
    
    return ret

def get_session_pids (user, target_proc="gnome-session"):
    p = subprocess.Popen (["ps -C %s -o pid=,user=" % (target_proc)],
                          stdout=subprocess.PIPE, close_fds=True, shell=True)
    pids = p.stdout.read ().strip ()
    p.wait ()

    ret = []
    for entry in pids.split ("\n"):
        pid, u = entry.strip ().split ()
        if u == user: 
            ret.append (pid)

    return ret

def run_in_user_session (user, cmd_args, target_proc="gnome-session"):
    for pid in get_session_pids (user, target_proc):
        new_env = get_env_from_pid (pid)
        z = subprocess.Popen (cmd_args, env=new_env, close_fds=True)
        z.wait ()

def notify_user (summary, body, user=None, expire=True):
    args = ["notify-send", "--icon=gtk-harddisk", "--urgency=normal", summary, body]
    if not expire:
        args.insert (3, "--expire-time=0")
    run_in_user_session (user if user else pwd.getpwuid (os.getuid ())[0], args)

def notify_user_err (summary, body, user=None, expire=False):
    args = ["notify-send", "--icon=gtk-harddisk", "--urgency=critical", summary, body]
    if not expire:
        args.insert (3, "--expire-time=0")
    run_in_user_session (user if user else pwd.getpwuid (os.getuid ())[0], args)

def get_manager_log_file_location ():
    return os.path.expanduser (os.path.join ('~', '.config', 'backup-manager', 'backup-manager.log')) 

def get_log_file_location ():
    return os.path.expanduser (os.path.join ('~', '.config', 'backup-manager', 'backend.log')) 

def get_config_location ():
    return os.path.expanduser (os.path.join ('~', '.config', 'backup-manager', 'config.json')) 

def get_file_list_location ():
    return os.path.expanduser (os.path.join ('~', '.config', 'backup-manager', 'file_list')) 

def get_logger (redirect=False):
    global logger_obj

    if logger_obj:
        return logger_obj
    
    lf = get_manager_log_file_location ()
    conf_dir = os.path.dirname (lf) 
    if not os.path.exists (conf_dir):
        os.makedirs (conf_dir, mode=0755)
        os.chmod (conf_dir, 0755)
    else:
        if os.path.exists (lf) and os.path.getsize (lf) > 10485760:
            os.unlink (lf)
    
    if redirect:
        os.close (2)

    logger_obj = logging.getLogger ("backup-manager")
    logger_obj.setLevel (logging.DEBUG)
    frm = logging.Formatter ("%(asctime)s %(filename)s: %(message)s", "%Y-%m-%d %H:%M")
    
    fh = logging.FileHandler (lf)
    fh.setLevel (logging.DEBUG)
    fh.setFormatter (frm)
    
    logger_obj.addHandler (fh)

    if redirect:
        os.close (1)
        os.dup (2)
    
    return logger_obj

def get_backup_manager ():
    import rdiff
    return rdiff.RdiffBackupManager ()

def schedule_backup (manager, config):
    try:
        pk = dbus.SessionBus ().get_object ("org.freedesktop.PolicyKit.AuthenticationAgent", "/")
        pk_agent = dbus.Interface (pk, "org.freedesktop.PolicyKit.AuthenticationAgent")
        pk_agent.ObtainAuthorization ("org.opensuse.backupmanager.schedule",
                                      dbus.UInt32 (0), dbus.UInt32 (os.getpid ()))
        
        bm = dbus.SystemBus ().get_object ("org.opensuse.BackupManager", "/org/opensuse/BackupManager")
        bm_if = dbus.Interface (bm, "org.opensuse.BackupManager")
        if config.backup_schedule == SCHEDULE_NEVER:
            bm_if.RemoveSchedule ()
        else:
            bm_if.AddSchedule ("%s %s" % (config.backup_schedule, config.backup_time))
    except dbus.DBusException, e:
        print >> sys.stderr, "PolicyKit failure: ", str (e)
        gsu = subprocess.Popen (["gnomesu", "%s" % (resource_filename (__name__, "service.py")),
                                "--", "--no-dbus", pwd.getpwuid (os.getuid ())[0],
                                "%s %s" % (config.backup_schedule, config.backup_time)],
                                stderr=subprocess.PIPE)
        gsu.wait ()
        if gsu.returncode != 0:
            raise BackupManagerException ("Failed to schedule backup: %s" % gsu.stderr.read ())

def block_suspend ():
    try:
        pm = dbus.SessionBus ().get_object ("org.freedesktop.PowerManagement",
                                            "/org/freedesktop/PowerManagement/Inhibit")
        pmi = dbus.Interface (pm, "org.freedesktop.PowerManagement.Inhibit")
        return pmi.Inhibit ("backup-manager", "Data is being saved.")
    except dbus.DBusException, e:
        raise BackupManagerException (str (e))

def unblock_suspend (cookie):
    try:
        pm = dbus.SessionBus ().get_object ("org.freedesktop.PowerManagement",
                                            "/org/freedesktop/PowerManagement/Inhibit")
        pmi = dbus.Interface (pm, "org.freedesktop.PowerManagement.Inhibit")
        pmi.UnInhibit (dbus.UInt32 (cookie))
    except dbus.DBusException, e:
        raise BackupManagerException (str (e))

def get_directory_size (dir):
    p = subprocess.Popen (["du", "-shkx", dir], stdout=subprocess.PIPE)
    output = p.stdout.read ()
    p.wait ()
    if p.returncode != 0:
        raise BackupManagerException ("Failed to get size of %s" % dir)

    fields = output.split ()
    return long (fields[0]) / 1024

def get_includes_size (config):
    ret = 0
    
    for inc in config.includes:
        if os.path.isdir (inc):
            ret = ret + get_directory_size (inc) * 1024
        else:
            ret = ret + os.path.getsize (inc) / 1024
   
    return ret / 1024

def get_mount_info (path):
    mp = None 
    mp_len = 0
    mp_dev = None
    mp_type = None

    fo = file ('/proc/mounts', 'r')
    for l in fo:
        if l.startswith ("rootfs"):
            continue

        fields = l.split (' ') 
        if path.startswith (fields[1]) and mp_len < len (fields[1]):
            mp = fields[1]
            mp_len = len (fields[1])
            mp_dev = fields[0]
            mp_type = fields[2]

    fo.close ()
    return mp, mp_dev, mp_type

def get_volume_info ():
    """Returns a hash of tuples (mount_point, label, device) keyed of uuid"""
    bus = dbus.SystemBus ()
    hal = bus.get_object ("org.freedesktop.Hal", "/org/freedesktop/Hal/Manager")
    manager = dbus.Interface (hal, "org.freedesktop.Hal.Manager")

    ret = {}
    for v in manager.FindDeviceByCapability ("volume"):
        dev_obj = bus.get_object ('org.freedesktop.Hal', v) 
        dev = dbus.Interface (dev_obj, "org.freedesktop.Hal.Device")
        uuid = dev.GetProperty ("volume.uuid")
        if uuid != "":
            parent_obj = bus.get_object ("org.freedesktop.Hal",
                                         dev.GetProperty ("info.parent")) 
            par = dbus.Interface (parent_obj, "org.freedesktop.Hal.Device")
            ret[uuid] = (dev.GetProperty ("volume.mount_point"),
                         dev.GetProperty ("volume.label"),
                         dev.GetProperty ("block.device"),
                         par.GetProperty ("storage.removable"))
    return ret 

def set_backup_path (config, path):
    mp, dev, typ = get_mount_info (path)
    if typ.find ("fuse") != -1:
        config.fuse_type = typ
        config.fuse_name = dev
        config.fuse_path = path
        config.remove ("backup_path")
        config.remove ("device_uuid")
    else:
        i = string.find (path, mp)
        config.backup_path = path[i + len (mp):]
        if config.backup_path.startswith ("/"):
            config.backup_path = config.backup_path[1:]
        config.device_uuid = uuid_from_device (dev)
        config.remove ("fuse_type")
        config.remove ("fuse_name")
        config.remove ("fuse_path")

def get_backup_path (config):
    if config.fuse_type:
        mp, dev, typ = get_mount_info (config.fuse_path)
        if typ != config.fuse_type:
            raise BackupManagerException ("FUSE backup path is not mounted")
        
        if dev != config.fuse_name:
            raise BackupManagerException ("FUSE backup path is using a different target")

        return config.fuse_path
    else:
        vol_info = get_volume_info ()
        if not vol_info.has_key (config.device_uuid):
            raise BackupManagerException ("Backup device is not attached")

        mp, label, dev, rem = vol_info[config.device_uuid]
        if not mp or mp == "":
            raise BackupManagerException ("Backup device is not mounted")

        return os.path.join (mp, config.backup_path)

def uuid_from_device (dev):
    vol_info = get_volume_info ()
    for k, v in vol_info.iteritems ():
        mp, label, device, rem = v
        if device == dev:
            return str (k)
