import os
import os.path
import re
import pwd
import string
import datetime
import subprocess
import gettext
from gettext import gettext as _
from backup_manager.core import *
from statvfs import *

class RdiffBackupManager (IBackupManager):
    def _generate_file_list (self, config, fl, src_location):
        if len (config.includes) == 0 and len (config.excludes) == 0:
            return False

        fl_dir = os.path.dirname (fl)
        if not os.path.exists (fl_dir):
            os.makedirs (fl_dir, mode=0755)
            os.chmod (fl_dir, 0755)
        
        fo = file (fl, 'w')
        
        for exc in config.excludes:
            if os.path.isdir (exc):
                fo.write ("- %s**\n" % exc)
            else:
                fo.write ("- %s\n" % exc)
        
        for inc in config.includes:
            if os.path.isdir (inc):
                fo.write ("+ %s**\n" % inc)
            else:
                fo.write ("+ %s\n" % inc)

        fo.write ("- %s/**" % src_location)
        fo.close ()
        return True

    def get_capabilities (self):
        return ["backup", "restore", "restore-specific", "purge", "restore-search"]

    def backup (self, config):
        self.verify_config (config)
        path = get_backup_path (config)
        hd = pwd.getpwuid (os.getuid ())[5]
        use_fl = self._generate_file_list (config, get_file_list_location (), hd)

        os.nice (19)
        args = ["rdiff-backup", "--force", "--verbosity=9", hd, path] if not use_fl else \
               ["rdiff-backup", "--force", "--verbosity=9", "--include-globbing-filelist", 
                get_file_list_location (), hd, path]
        
        lf = get_log_file_location ()
        fo = file (lf, "w")

        print "running", args
        rdp = subprocess.Popen (args, close_fds=True, stdout=fo, stderr=fo)
        subprocess.call (["ionice", "-n7", "-p%s" % (rdp.pid)])
        rdp.wait ()
        
        fo.flush ()
        fo.close ()
        
        if rdp.returncode != 0:
            raise BackupManagerException (_("Backup failed.  See %s for details.") % lf)
        else:
            config.last_backup = time.time ()
            config.save ()

        self.purge_old_data (config)
    
    def verify_config (self, config):
        if config.backup_schedule == SCHEDULE_NEVER:
            return True

        if subprocess.call (["rdiff-backup", "--version"]) != 0:
            raise BackupManagerException (_("Failed to find rdiff-backup"))
        
        path = get_backup_path (config)
        hd = pwd.getpwuid (os.getuid ())[5]
        hd_mp, hd_dev, hd_type = get_mount_info (os.path.realpath (hd))
        bp_mp, bp_dev, bp_type = get_mount_info (os.path.realpath (path))
        
        if hd_mp == None:
            raise BackupManagerException (_("Unable to get mount point for %s") % hd)
        
        if bp_mp == None:
            raise BackupManagerException (_("Unable to get mount point for %s") % path)
        
        if hd_dev == bp_dev:
            raise BackupManagerException (_("'%s' and '%s' can't be on the same device") % (hd, path))
        
        if not os.access (path, os.R_OK | os.W_OK):
            raise BackupManagerException (_("You don't have read-write access to %s") % path)
        
        if not os.path.exists (os.path.join (hd, "rdiff-backup-data")):
            dest_info = os.statvfs (path)
            dest_size = dest_info[F_BAVAIL] * dest_info[F_FRSIZE] / 1048576
            if config.file_selection == "all":
                src_size = get_directory_size (hd)
            else:
                src_size = get_includes_size (config)

            if src_size > dest_size:
                raise BackupManagerException (_("Your data is %sMB, but %s only has %sMB available") % \
                                       (src_size, path, dest_size))
        return True
    
    def verify_restore_data (self, src, dest, files, time):
        if not os.path.exists (os.path.join (src, "rdiff-backup-data")):
            raise BackupManagerException (_("%s is not a valid rdiff-backup directory") % src)
        
        if not os.access (dest, os.R_OK | os.W_OK):
            raise BackupManagerException (_("You don't have read-write access to %s") % dest)
        
        if time > datetime.datetime.now ():
            raise BackupManagerException (_("You can't restore to a time in the future"))
        
        if src == dest:
            raise BackupManagerException (_("Source and destination can't be the same."))

        if files:
            for f in files:
                if type (f) == tuple:
                    if not f[0].startswith (src):
                        raise BackupManagerException (_("%s is not rooted at %s") % (f[0], src))
                else:
                    if not f.startswith (src):
                        raise BackupManagerException (_("%s is not rooted at %s") % (f, src))
                
        return True

    def _convert_path_from_escaped (self, path):
        return re.sub (';[0-9]{3}', lambda m: chr (int (m.group (0)[1:])), path)

    def _convert_path_to_escaped (self, path):
        return re.sub ('[A-Z]', lambda m: ";%03d" % ord (m.group (0)[0]), path)

    def restore (self, src, dest, time):
        os.nice (19)
        rp = time.isoformat () if time else datetime.now.isoformat ()
        args = ["rdiff-backup", "-r", rp, self._convert_path_from_escaped (src), dest] 
        
        lf = get_log_file_location ()
        fo = file (lf, "w")

        print "running", args
        rdp = subprocess.Popen (args, close_fds=True, stdout=fo, stderr=fo)
        subprocess.call (["ionice", "-n7", "-p%s" % (rdp.pid)])
        rdp.wait ()
        
        fo.flush ()
        fo.close ()
        
        if rdp.returncode != 0:
            raise BackupManagerException ("Restore failed.  See %s for details." % lf)
    
    def restore_specific  (self, src, dest, files):
        os.nice (19)
        src_len = len (self._convert_path_from_escaped (src))
        
        for restore_file, time in files:
            conv_file = self._convert_path_from_escaped (restore_file)
            
            if not time:
                restore_point = "now" 
            elif type (time) == datetime.datetime:
                restore_point = time.isoformat ()
            else:
                restore_point = time
            
            df = dest + os.path.sep + conv_file[src_len:]
            df_dir = os.path.dirname (df)

            if not os.path.exists (df_dir):
                os.makedirs (df_dir, mode=0755)
                os.chmod (df_dir, 0755)
            
            args = ["rdiff-backup", "--force", "-r", restore_point, conv_file, df] 
            lf = get_log_file_location ()
            fo = file (lf, "w")

            print "running", args
            rdp = subprocess.Popen (args, close_fds=True, stdout=fo, stderr=fo)
            subprocess.call (["ionice", "-n7", "-p%s" % (rdp.pid)])
            rdp.wait ()
            
            fo.flush ()
            fo.close ()
            
            if rdp.returncode != 0:
                raise BackupManagerException ("Restore failed.  See %s for details." % lf)

    def purge_old_data (self, config):
        if config.purge_days == None or config.purge_days == 0:
            return

        self.verify_config (config)
        path = get_backup_path (config)

        os.nice (19)
        args = ["rdiff-backup", "--remove-older-than", "%dD" % config.purge_days, path] 
        
        lf = get_log_file_location ()
        fo = file (lf, "w")

        print "running", args
        rdp = subprocess.Popen (args, close_fds=True, stdout=fo, stderr=fo)
        subprocess.call (["ionice", "-n7", "-p%s" % (rdp.pid)])
        rdp.wait ()
        
        fo.flush ()
        fo.close ()
        
        if rdp.returncode != 0:
            raise BackupManagerException (_("Data purge failed.  See %s for details.") % lf)

    def restore_search (self, config, key):
        """Returns a list of tuples (pretty name, name, time_string)"""
        self.verify_config (config)
        ckey = self._convert_path_to_escaped (key)
        path = get_backup_path (config)
        if not path.endswith ("/"):
            path = path + "/"

        p = subprocess.Popen (["find", path, "-name", "*%s*" % ckey], stdout=subprocess.PIPE)
        output = p.stdout.read ()
        p.wait ()

        output = output.strip ()
        if output == "" or output == "\n":
            return None

        ret = []
        for hit in output.split ("\n"):
            if hit == "" or hit == "\n" or hit.endswith (".missing") or \
               os.path.dirname (hit).endswith ("rdiff-backup-data"):
                continue

            if hit.startswith (path + "rdiff-backup-data/increments"):
                fields = hit.split (".")
                fields.reverse ()
                if fields[0] not in ["dir", "gz", "diff", "snapshot"]:
                    continue

                conv_hit = self._convert_path_from_escaped (hit[len (path + "/rdiff-backup-data/increments"):])
                fields = conv_hit.split (".")
                fields.reverse ()
                time_field = None
                
                for f in fields:
                    if f not in ["gz", "dir", "diff", "snapshot"]:
                        i = string.find (conv_hit, f)
                        conv_hit = "%s (from %s)" % (conv_hit[0:i - 1], f)
                        ret.append ((path + conv_hit, path + conv_hit[0:i - 1], f))
                        break
            else:
                conv_hit = self._convert_path_from_escaped (hit[len (path):])
                ret.append ((path + conv_hit, path + conv_hit, None))
        
        return ret
