#!/usr/bin/env python

import os
import pwd
import sys
import tempfile
import subprocess
import random
import dbus
import dbus.service
import dbus.mainloop.glib
import gobject
from optparse import OptionParser
from backup_manager.core import *

def string_to_cron_schedule (sched_str):
    fields = sched_str.split ()
    schedule = int (fields[0])
    cron_str, hour, min = None, None, None
    
    if len (fields) == 2:
        time_fields = fields[1].split (":")
        hour = int (time_fields[0]) % 24
        min = int (time_fields[1])
        if time_fields[2] == "PM":
            hour = hour + 12
        
    if schedule == SCHEDULE_NEVER:
        return None
    elif schedule == SCHEDULE_HOURLY:
        cron_str = "0 * * * *" 
    elif schedule == SCHEDULE_DAILY:
        cron_str = "%d %d * * *" % (min, hour) if hour else "0 0 * * *"
    elif schedule == SCHEDULE_WEEKLY:
        day = random.randint (0, 6)
        cron_str = "%d %d * * %d" % (min, hour, day) if hour else "0 0 * * %d" % day
    elif schedule == SCHEDULE_MONTHLY:
        cron_str = "0 0 %d * *" % random.randint (1, 28) 
    else:
        raise BackupManagerServiceException ("Invalid schedule %s" % schedule)

    return cron_str + " backup-manager --now > /dev/null 2>&1\n" 

def _get_cron_contents (user):
    prog = subprocess.Popen (["crontab", "-u", user, "-l"],
                             stdout=subprocess.PIPE,
                             close_fds=True, cwd='/')
    lines = prog.stdout.readlines ()
    prog.wait ()
    return lines

def _set_cron_contents (user, lines):
    (fd, fname) = tempfile.mkstemp ()
    fo = os.fdopen (fd, "w")
    
    for l in lines:
        if l.startswith ("#"):
            continue
        
        fo.write (l)

    fo.close ()
   
    prog2 = subprocess.Popen (["crontab", "-u", user, fname],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              close_fds=True, cwd='/')
    prog2.wait ()
    os.unlink (fname)

    if prog2.returncode != 0:
        raise BackupManagerServiceException ("Failed to setup crontab")

def remove_schedule (user):
    lines = []
    for l in _get_cron_contents (user):
        if l.find ("backup-manager") == -1:
            lines.append (l)

    _set_cron_contents (user, lines)

def add_schedule (user, new_sched):
    ht = {}
    for l in _get_cron_contents (user):
        ht[l] = l

    lines = []
    for k, v in ht.iteritems ():
        if v.find ("backup-manager") != -1:
            continue
        
        lines.append (v)
    
    sch = string_to_cron_schedule (new_sched)
    if sch:
        lines.append (sch)
    
    # periodically check to see if we've missed backups
    lines.append ("0 3,12,21 * * * backup-manager --check > /dev/null 2>&1\n")
    _set_cron_contents (user, lines)

class BackupManagerServiceException (dbus.DBusException):
    _dbus_error_name = "org.opensuse.BackupManagerServiceException"

class BackupManagerService (dbus.service.Object):
    def __init__ (self):
        bn = dbus.service.BusName ("org.opensuse.BackupManager", bus=dbus.SystemBus ())
        dbus.service.Object.__init__ (self, bn, "/org/opensuse/BackupManager")

    def _get_user_if_authorized (self, sender):
        if not sender:
            raise BackupManagerServiceException ("Failed to get sender")
        
        sys_bus = dbus.SystemBus ()
        dbus_service = sys_bus.get_object ("org.freedesktop.DBus", "/")
        dbus_if = dbus.Interface (dbus_service, "org.freedesktop.DBus")
        user_id = dbus_if.GetConnectionUnixUser (sender)

        pk_service = sys_bus.get_object ("org.freedesktop.PolicyKit", "/")
        pk = dbus.Interface (pk_service, "org.freedesktop.PolicyKit")
        authed = pk.IsSystemBusNameAuthorized ("org.opensuse.backupmanager.schedule", sender, False)
        if authed != "yes":
            raise BackupManagerServiceException ("Not authorized")
        
        return pwd.getpwuid (user_id)[0]

    @dbus.service.method ("org.opensuse.BackupManager", in_signature="", out_signature="", sender_keyword="sender")
    def RemoveSchedule (self, sender=None):
        user = self._get_user_if_authorized (sender)
        remove_schedule (user)
    
    @dbus.service.method ("org.opensuse.BackupManager", in_signature="s", out_signature="", sender_keyword="sender")
    def AddSchedule (self, schedule, sender=None):
        user = self._get_user_if_authorized (sender)
        add_schedule (user, schedule)

def main ():
    parser = OptionParser (usage="%prog [options] user schedule")
    parser.add_option ("-n", "--no-dbus", dest="nodbus", action="store_true",
                       default=None, help="Don't start as a dbus service")
    (options, args) = parser.parse_args ()
    
    if options.nodbus:
        if len (args) < 2:
            parser.error ("Invalid usage")

        try:
            user = args[0]
            sched = args[1]
            pwd.getpwnam (user)

            if sched.startswith (str (SCHEDULE_NEVER)):
                remove_schedule (user)
            else:
                add_schedule (user, sched)
        except Exception, e:
            parser.error (str (e))
    else:
        dbus.mainloop.glib.DBusGMainLoop (set_as_default=True)
        bm = BackupManagerService ()
        loop = gobject.MainLoop ()
        loop.run ()

if __name__ == "__main__":
    main ()
